import { HPCCElement } from "./element";
const _allMeta = new WeakMap();
const _allMetaByName = {};
function initMeta(target) {
    let retVal;
    if (!_allMeta.has(target)) {
        retVal = {
            tag: "",
            template: { html: "", directives: [] },
            styles: "",
            properties: [],
            observedAttributes: [],
            observedProperties: [],
            observed: {},
            events: []
        };
        _allMeta.set(target, retVal);
        _allMetaByName[target.name] = retVal;
    }
    else {
        retVal = _allMeta.get(target);
    }
    return retVal;
}
export function allMeta() {
    return _allMetaByName;
}
export function classMeta(target) {
    if (typeof target === "string") {
        return _allMetaByName[target];
    }
    return _allMeta.get(target);
}
export function instanceMeta(target) {
    return initMeta(Object.getPrototypeOf(target).constructor);
}
export function customElement(name, opts) {
    const { template = { html: "", directives: [] }, styles = "", events = [] } = opts || {};
    function decorator(target) {
        const meta = initMeta(target);
        //  Gather inherited meta  ---
        let allProperties = [];
        let self = target;
        while (true) {
            allProperties = allProperties.concat(initMeta(self).properties);
            if (self === HPCCElement) {
                break;
            }
            self = Object.getPrototypeOf(self);
        }
        meta.tag = name;
        meta.template = template;
        meta.styles = styles;
        meta.observedAttributes = allProperties
            .filter(prop => prop.isAttribute)
            .map(prop => prop.name);
        meta.observedProperties = allProperties
            .filter(prop => !prop.isAttribute)
            .map(prop => prop.name);
        meta.observed = {};
        meta.events = events;
        allProperties.forEach(prop => meta.observed[prop.name] = prop);
        customElements?.define(name, target);
        return;
    }
    return decorator;
}
function changedHandler(target, opts) {
    if (opts.name !== opts.name.toLowerCase()) {
        throw new Error(`Attributes must be lowercase: ${target.constructor.name}.${opts.name}`);
    }
    const meta = initMeta(target.constructor);
    meta.properties.push(opts);
    const innerID = `_${opts.name}`;
    Object.defineProperty(target, opts.name, {
        set: function (newVal) {
            const oldVal = this[innerID];
            this[innerID] = newVal;
            this._fire(opts.name, oldVal, newVal);
            if (opts.isAttribute) {
                if (this.isConnected) {
                    this.attr(opts.name, newVal);
                }
            }
        },
        get: function () { return this[innerID]; }
    });
}
function attrProp(isAttribute, configOrTarget, prop) {
    const options = {
        name: prop ?? "",
        type: "string",
        isAttribute: isAttribute,
        ...(prop === undefined ? configOrTarget : {})
    };
    function decorator($target, $prop) {
        changedHandler($target, { ...options, name: $prop });
    }
    if (prop === undefined) {
        return decorator;
    }
    decorator(configOrTarget, prop);
}
export function attribute(configOrTarget, prop) {
    return attrProp(true, configOrTarget, prop);
}
export function property(configOrTarget, prop) {
    return attrProp(false, configOrTarget, prop);
}
//# sourceMappingURL=decorator.js.map