(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global["@hpcc-js/wc-core"] = {}));
})(this, (function (exports) { 'use strict';

    function css(templateData, ...args) {
      let retVal = templateData[0];

      for (let i = 0; i < args.length; ++i) {
        retVal += String(args[i]);
        retVal += templateData[i + 1];
      }

      return retVal;
    }
    function display(display) {
      return `\
:host([hidden]) {
    display: none
}

:host {
    display: ${display}
}`;
    }

    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */
    function __decorate(decorators, target, key, desc) {
      var c = arguments.length,
          r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc,
          d;
      if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
      return c > 3 && r && Object.defineProperty(target, key, r), r;
    }

    var root = typeof globalThis !== "undefined" ? globalThis : window;
    var isNode = typeof process !== "undefined" && process.versions != null && process.versions.node != null;
    isNode && process.env != null && (process.env.TRAVIS != null || process.env.GITHUB_ACTIONS != null);

    var requestAnimationFrame; // let cancelAnimationFrame: CancelAnimationFrame;

    (function () {
      if (root.requestAnimationFrame) {
        requestAnimationFrame = root.requestAnimationFrame; // cancelAnimationFrame = root.cancelAnimationFrame;
      } else {
        var lastTime_1 = 0;

        requestAnimationFrame = function (callback) {
          var currTime = new Date().getTime();
          var timeToCall = Math.max(0, 16 - (currTime - lastTime_1));
          var id = setTimeout(function () {
            return callback(currTime + timeToCall);
          }, timeToCall);
          lastTime_1 = currTime + timeToCall;
          return id;
        }; // cancelAnimationFrame = function (handle: number): void {
        //     clearTimeout(handle);
        // };

      }
    })();

    var Message =
    /** @class */
    function () {
      function Message() {}

      Object.defineProperty(Message.prototype, "canConflate", {
        get: function () {
          return false;
        },
        enumerable: false,
        configurable: true
      });

      Message.prototype.conflate = function (other) {
        return false;
      };

      Message.prototype.void = function () {
        return false;
      };

      return Message;
    }();

    var Dispatch =
    /** @class */
    function () {
      function Dispatch() {
        this._observerID = 0;
        this._observers = [];
        this._messageBuffer = [];
      }

      Dispatch.prototype.observers = function () {
        return this._observers;
      };

      Dispatch.prototype.messages = function () {
        var retVal = [];

        this._messageBuffer.forEach(function (msg) {
          if (!retVal.some(function (msg2) {
            return msg2.canConflate && msg2.conflate(msg);
          })) {
            retVal.push(msg);
          }
        });

        return retVal;
      };

      Dispatch.prototype.dispatchAll = function () {
        this.dispatch(this.messages());
        this.flush();
      };

      Dispatch.prototype.dispatch = function (messages) {
        if (messages.length === 0) return;
        this.observers().forEach(function (o) {
          var msgs = messages.filter(function (m) {
            return !m.void() && (o.type === undefined || m instanceof o.type);
          });

          if (msgs.length) {
            o.callback(msgs);
          }
        });
      };

      Dispatch.prototype.hasObserver = function () {
        return this._observers.length > 0;
      };

      Dispatch.prototype.flush = function () {
        this._messageBuffer = [];
      };

      Dispatch.prototype.send = function (msg) {
        this.dispatch([msg]);
      };

      Dispatch.prototype.post = function (msg) {
        var _this = this;

        this._messageBuffer.push(msg);

        requestAnimationFrame(function () {
          return _this.dispatchAll();
        });
      };

      Dispatch.prototype.attach = function (callback, type) {
        var context = this;
        var id = ++this._observerID;

        this._observers.push({
          id: id,
          type: type,
          callback: callback
        });

        return {
          release: function () {
            context._observers = context._observers.filter(function (o) {
              return o.id !== id;
            });
          },
          unwatch: function () {
            this.release();
          }
        };
      };

      return Dispatch;
    }();

    class Directive {}
    class Ref extends Directive {
      constructor(propertyName) {
        super();
        this.propertyName = propertyName;
      }

    }
    function html(strings, ...values) {
      let html = strings[0];
      const directives = [];

      for (let i = 0; i < values.length; ++i) {
        const arg = values[i];

        if (arg instanceof Directive) {
          directives.push(arg);

          if (arg instanceof Ref) {
            html += `id="${arg.propertyName}"`;
          } else {
            html += String(values[i]);
          }
        } else {
          throw new Error("Invalid Directive");
        }

        html += strings[i + 1];
      }

      return {
        html,
        directives
      };
    }
    function ref(propertyName) {
      return new Ref(propertyName);
    }

    class AttrChangedMessage extends Message {
      constructor(what, oldValue, newValue) {
        super();
        this.changes = {};
        this.changes[what] = {
          oldValue,
          newValue
        };
      }

      get canConflate() {
        return true;
      }

      conflate(other) {
        for (const what in other.changes) {
          const thisChange = this.changes[what];
          const otherChange = other.changes[what];

          if (thisChange) {
            this.changes[what].newValue = otherChange?.newValue;
          } else {
            this.changes[what] = otherChange;
          }
        }

        return true;
      }

    }

    const DefaultEventOptions = {
      bubbles: true,
      composed: true,
      cancelable: true
    };
    class HPCCElement extends HTMLElement {
      constructor() {
        super();
        this.$meta = instanceMeta(this);
        this.$dispatch = new Dispatch();

        this._fire = (what, oldVal = false, newVal = true) => {
          this.$dispatch.post(new AttrChangedMessage(what, oldVal, newVal));
        };

        this.$_initialized = false; //  Gather user values set prior to "upgrade"  ---

        this.$upgradeProperties(); //  Initialize shadow DOM  ---

        this.attachShadow({
          mode: "open"
        });
        this.shadowRoot.innerHTML = this.$meta.template?.html.trim() || "";

        for (const directive of this.$meta.template?.directives || []) {
          if (directive instanceof Ref) {
            const ref = this.shadowRoot.getElementById(String(directive.propertyName));
            this[directive.propertyName] = ref;
            ref?.removeAttribute("id");
          }
        }

        this.$styles = document.createElement("style");
        this.$styles.innerHTML = this.$meta.styles.trim();
        this.shadowRoot.insertBefore(this.$styles, this.shadowRoot.firstChild);
      }

      static register() {//  Do nothing - calling this will ensure WebPack / RollupJS won't "Tree Shake" it out of the final bundle  ---
      }

      static get observedAttributes() {
        return classMeta(this).observedAttributes;
      }

      set styles(_) {
        this.$styles.innerHTML = _;
      }

      get styles() {
        return this.$styles.innerHTML;
      }

      $upgradeProperties() {
        Object.keys(this.$meta.observed).forEach(prop => {
          const userValueID = `__${prop}`;

          if (this.hasOwnProperty(prop)) {
            this[userValueID] = this[prop];
            delete this[prop];
          }
        });
      }

      coercedValue(qualifiedName, value) {
        switch (this.$meta.observed[qualifiedName].type) {
          case "boolean":
            return !!value;

          case "number":
            const retVal = Number(value);
            return isNaN(retVal) ? null : retVal;

          case "string":
          default:
            return value;
        }
      }

      attr(qualifiedName, _) {
        if (_ === undefined) {
          switch (this.$meta.observed[qualifiedName].type) {
            case "boolean":
              return this.coercedValue(qualifiedName, this.hasAttribute(qualifiedName) ? "true" : "");

            case "number":
            case "string":
            default:
              return this.coercedValue(qualifiedName, this.getAttribute(qualifiedName));
          }
        } else {
          switch (this.$meta.observed[qualifiedName].type) {
            case "boolean":
              if (_) {
                this.setAttribute(qualifiedName, "true");
              } else {
                this.removeAttribute(qualifiedName);
              }

              break;

            case "number":
              this.setAttribute(qualifiedName, _.toString());
              break;

            case "string":
            default:
              this.setAttribute(qualifiedName, _);
          }
        }

        return this;
      }

      initalizeAttributes() {
        if (this.$_initialized) return {};
        this.$_initialized = true;
        const retVal = {};
        this.$meta.observedAttributes.forEach(attr => {
          const userValueID = `__${attr}`;
          const innerID = `_${attr}`;

          if (this[userValueID] !== undefined) {
            this[innerID] = this[userValueID];
            delete this[userValueID];
          }

          this.attr(attr, this[innerID]);
          retVal[attr] = {
            oldValue: undefined,
            newValue: this[innerID]
          };
        });
        this.$meta.observedProperties.forEach(prop => {
          const userValueID = `__${prop}`;
          const innerID = `_${prop}`;

          if (this[userValueID] !== undefined) {
            this[innerID] = this[userValueID];
            delete this[userValueID];
          }

          retVal[prop] = {
            oldValue: undefined,
            newValue: this[innerID]
          };
        });
        return retVal;
      }

      connectedCallback() {
        const changes = this.initalizeAttributes();
        this.enter();
        this.update(changes);
        this.$dispatch.flush();
        this.$dispatchHandle = this.$dispatch.attach(messages => {
          if (this.isConnected) {
            const changes = {};
            if (messages.length > 1) throw new Error("Conflation issue.");

            for (const what in messages[0].changes) {
              const change = messages[0].changes[what];

              if (change?.oldValue !== change?.newValue) {
                changes[what] = change;
              }
            }

            if (Object.keys(changes).length) {
              this.update(changes);
            }
          }
        });
      }

      disconnectedCallback() {
        this.$dispatchHandle.release();
        this.exit();
      }

      adoptedCallback() {}

      attributeChangedCallback(name, _oldValue, newValue) {
        const innerID = `_${name}`;
        const coercedValue = this.coercedValue(name, newValue);

        if (this[innerID] !== coercedValue) {
          const oldValue = this[innerID];
          this[innerID] = coercedValue;

          this._fire(name, oldValue, coercedValue);
        }
      } //  Lifecycle  ---


      enter() {
        //  Debugging, remove for production  ---
        for (const key of this.$meta.observedAttributes) {
          if (this[key] !== this.attr(key)) {
            console.log("enter sync error", key, this[key], this.attr(key));
          }
        } //  Debugging, remove for production  ---

      }

      update(_changes) {
        //  Debugging, remove for production  ---
        for (const key of this.$meta.observedAttributes) {
          if (this[key] !== this.attr(key)) {
            console.log("update sync error", key, this[key], this.attr(key));
          }
        } //  Debugging, remove for production  ---

      }

      exit() {} //  Events  ---


      $emit(type, detail, options) {
        const opts = {
          detail,
          ...DefaultEventOptions,
          ...options
        };

        if (this.isConnected) {
          return this.dispatchEvent(new CustomEvent(type, opts));
        }

        return opts.cancelable ? true : false;
      }

    }

    const _allMeta = new WeakMap();

    const _allMetaByName = {};

    function initMeta(target) {
      let retVal;

      if (!_allMeta.has(target)) {
        retVal = {
          tag: "",
          template: {
            html: "",
            directives: []
          },
          styles: "",
          properties: [],
          observedAttributes: [],
          observedProperties: [],
          observed: {},
          events: []
        };

        _allMeta.set(target, retVal);

        _allMetaByName[target.name] = retVal;
      } else {
        retVal = _allMeta.get(target);
      }

      return retVal;
    }

    function allMeta() {
      return _allMetaByName;
    }
    function classMeta(target) {
      if (typeof target === "string") {
        return _allMetaByName[target];
      }

      return _allMeta.get(target);
    }
    function instanceMeta(target) {
      return initMeta(Object.getPrototypeOf(target).constructor);
    }
    function customElement(name, opts) {
      const {
        template = {
          html: "",
          directives: []
        },
        styles = "",
        events = []
      } = opts || {};

      function decorator(target) {
        const meta = initMeta(target); //  Gather inherited meta  ---

        let allProperties = [];
        let self = target;

        while (true) {
          allProperties = allProperties.concat(initMeta(self).properties);

          if (self === HPCCElement) {
            break;
          }

          self = Object.getPrototypeOf(self);
        }

        meta.tag = name;
        meta.template = template;
        meta.styles = styles;
        meta.observedAttributes = allProperties.filter(prop => prop.isAttribute).map(prop => prop.name);
        meta.observedProperties = allProperties.filter(prop => !prop.isAttribute).map(prop => prop.name);
        meta.observed = {};
        meta.events = events;
        allProperties.forEach(prop => meta.observed[prop.name] = prop);
        customElements?.define(name, target);
        return;
      }

      return decorator;
    }

    function changedHandler(target, opts) {
      if (opts.name !== opts.name.toLowerCase()) {
        throw new Error(`Attributes must be lowercase: ${target.constructor.name}.${opts.name}`);
      }

      const meta = initMeta(target.constructor);
      meta.properties.push(opts);
      const innerID = `_${opts.name}`;
      Object.defineProperty(target, opts.name, {
        set: function (newVal) {
          const oldVal = this[innerID];
          this[innerID] = newVal;

          this._fire(opts.name, oldVal, newVal);

          if (opts.isAttribute) {
            if (this.isConnected) {
              this.attr(opts.name, newVal);
            }
          }
        },
        get: function () {
          return this[innerID];
        }
      });
    }

    function attrProp(isAttribute, configOrTarget, prop) {
      const options = {
        name: prop ?? "",
        type: "string",
        isAttribute: isAttribute,
        ...(prop === undefined ? configOrTarget : {})
      };

      function decorator($target, $prop) {
        changedHandler($target, { ...options,
          name: $prop
        });
      }

      if (prop === undefined) {
        return decorator;
      }

      decorator(configOrTarget, prop);
    }

    function attribute(configOrTarget, prop) {
      return attrProp(true, configOrTarget, prop);
    }
    function property(configOrTarget, prop) {
      return attrProp(false, configOrTarget, prop);
    }

    class HPCCResizeElement extends HPCCElement {
      constructor() {
        super(...arguments);
        this._prevSize = {
          width: 0,
          height: 0
        };
        this.observer = new ResizeObserver(() => {
          const size = {
            width: this.clientWidth,
            height: this.clientHeight
          };

          this._fire("resize", this._prevSize, size);

          this._prevSize = size;
        });
      }

      connectedCallback() {
        super.connectedCallback();
        this.observer.observe(this);
      }

      disconnectedCallback() {
        this.observer.unobserve(this);
        super.disconnectedCallback();
      }

    }

    const template = html`\
`;
    const styles = css`${display("inline-block")}:host{}:host > div{background-color: pink}:host > div > div{position: absolute;background-color: palegreen}`;
    exports.HPCCPanelElement = class HPCCPanelElement extends HPCCResizeElement {
      constructor() {
        super();
      }

      enter() {
        super.enter();
      }

      update(changes) {
        super.update(changes);
      }

    };
    exports.HPCCPanelElement = __decorate([customElement("hpcc-panel", {
      template,
      styles
    })], exports.HPCCPanelElement);

    const isTrue = value => value === true || value === 1 || value === "true" || value === "1";

    class HPCCSVGElement extends HPCCResizeElement {
      constructor() {
        super();

        if (!this._svg) {
          throw new Error("SVG element not found");
        }
      }

      enter() {}

      update(changes) {
        super.update(changes);

        if ("resize" in changes) {
          this._svg.setAttribute("width", `${this.clientWidth}px`);

          this._svg.setAttribute("height", `${this.clientHeight}px`);

          this._svg.setAttribute("viewBox", `0 0 ${this._svg.clientWidth} ${this._svg.clientHeight}`);
        }
      }

    }

    exports.AttrChangedMessage = AttrChangedMessage;
    exports.DefaultEventOptions = DefaultEventOptions;
    exports.Directive = Directive;
    exports.HPCCElement = HPCCElement;
    exports.HPCCResizeElement = HPCCResizeElement;
    exports.HPCCSVGElement = HPCCSVGElement;
    exports.Ref = Ref;
    exports.allMeta = allMeta;
    exports.attribute = attribute;
    exports.classMeta = classMeta;
    exports.css = css;
    exports.customElement = customElement;
    exports.display = display;
    exports.html = html;
    exports.instanceMeta = instanceMeta;
    exports.isTrue = isTrue;
    exports.property = property;
    exports.ref = ref;

    Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=index.umd.js.map
